@@ -25,11 +25,11 @@ except ImportError:
 
     or via conda using
 
-     conda install pandas matplotlib datashader bokeh holoviews colorcet
+     conda install pandas matplotlib datashader bokeh holoviews colorcet scikit-image
     """
     )
     raise ImportError(
-        "umap.plot requires pandas matplotlib datashader bokeh holoviews and colorcet to be "
+        "umap.plot requires pandas matplotlib datashader bokeh holoviews scikit-image and colorcet to be "
         "installed"
     ) from None
 
@@ -39,10 +39,10 @@ import sklearn.neighbors
 
 from matplotlib.patches import Patch
 
-from umap.nndescent import initialise_search, initialized_nnd_search
-from umap.utils import deheap_sort, submatrix
+from umap.utils import submatrix, average_nn_distance
 
 from bokeh.plotting import show as show_interactive
+from bokeh.plotting import output_file, output_notebook
 from bokeh.layouts import column
 from bokeh.models import CustomJS, TextInput
 from matplotlib.pyplot import show as show_static
@@ -63,11 +63,11 @@ darkpurple_cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
     "darkpurple", colorcet.linear_bmw_5_95_c89
 )
 
-plt.register_cmap("fire", fire_cmap)
-plt.register_cmap("darkblue", darkblue_cmap)
-plt.register_cmap("darkgreen", darkgreen_cmap)
-plt.register_cmap("darkred", darkred_cmap)
-plt.register_cmap("darkpurple", darkpurple_cmap)
+plt.colormaps.register(fire_cmap, name="fire")
+plt.colormaps.register(darkblue_cmap, name="darkblue")
+plt.colormaps.register(darkgreen_cmap, name="darkgreen")
+plt.colormaps.register(darkred_cmap, name="darkred")
+plt.colormaps.register(darkpurple_cmap, name="darkpurple")
 
 
 def _to_hex(arr):
@@ -149,6 +149,31 @@ _themes = {
 _diagnostic_types = np.array(["pca", "ica", "vq", "local_dim", "neighborhood"])
 
 
+def _get_embedding(umap_object):
+    if hasattr(umap_object, "embedding_"):
+        return umap_object.embedding_
+    elif hasattr(umap_object, "embedding"):
+        return umap_object.embedding
+    else:
+        raise ValueError("Could not find embedding attribute of umap_object")
+
+
+def _get_metric(umap_object):
+    if hasattr(umap_object, "metric"):
+        return umap_object.metric
+    else:
+        # Assume euclidean if no attribute per cuML.UMAP
+        return "euclidean"
+
+
+def _get_metric_kwds(umap_object):
+    if hasattr(umap_object, "_metric_kwds"):
+        return umap_object._metric_kwds
+    else:
+        # Assume no keywords exist
+        return {}
+
+
 def _embed_datashader_in_an_axis(datashader_image, ax):
     img_rev = datashader_image.data[::-1]
     mpl_img = np.dstack([_blue(img_rev), _green(img_rev), _red(img_rev)])
@@ -157,7 +182,7 @@ def _embed_datashader_in_an_axis(datashader_image, ax):
 
 
 def _nhood_search(umap_object, nhood_size):
-    if umap_object._small_data:
+    if hasattr(umap_object, "_small_data") and umap_object._small_data:
         dmat = sklearn.metrics.pairwise_distances(umap_object._raw_data)
         indices = np.argpartition(dmat, nhood_size)[:, :nhood_size]
         dmat_shortened = submatrix(dmat, indices, nhood_size)
@@ -167,49 +192,22 @@ def _nhood_search(umap_object, nhood_size):
     else:
         rng_state = np.empty(3, dtype=np.int64)
 
-        if len(umap_object._metric_kwds) >= 1:
-            _dist = umap_object._input_distance_func
-            _args = tuple(umap_object._metric_kwds.values())
-
-            @numba.njit()
-            def _metric(x, y):
-                _dist(x, y, *_args)
-
-        else:
-            _metric = umap_object._input_distance_func
-
-        init = initialise_search(
-            umap_object._rp_forest,
+        indices, dists = umap_object._knn_search_index.query(
             umap_object._raw_data,
-            umap_object._raw_data,
-            int(nhood_size * umap_object.transform_queue_size),
-            rng_state,
-            _metric,
+            k=nhood_size,
         )
 
-        result = initialized_nnd_search(
-            umap_object._raw_data,
-            umap_object._search_graph.indptr,
-            umap_object._search_graph.indices,
-            init,
-            umap_object._raw_data,
-            _metric,
-        )
-
-        indices, dists = deheap_sort(result)
-        indices = indices[:, :nhood_size]
-        dists = dists[:, :nhood_size]
-
     return indices, dists
 
 
-@numba.jit()
+@numba.jit(nopython=False)
 def _nhood_compare(indices_left, indices_right):
     """Compute Jaccard index of two neighborhoods"""
     result = np.empty(indices_left.shape[0])
 
     for i in range(indices_left.shape[0]):
-        intersection_size = np.intersect1d(indices_left[i], indices_right[i]).shape[0]
+        intersection_size = np.intersect1d(indices_left[i], indices_right[i], 
+                                           assume_unique=True).shape[0]
         union_size = np.unique(np.hstack([indices_left[i], indices_right[i]])).shape[0]
         result[i] = float(intersection_size) / float(union_size)
 
@@ -218,10 +216,10 @@ def _nhood_compare(indices_left, indices_right):
 
 def _get_extent(points):
     """Compute bounds on a space with appropriate padding"""
-    min_x = np.min(points[:, 0])
-    max_x = np.max(points[:, 0])
-    min_y = np.min(points[:, 1])
-    max_y = np.max(points[:, 1])
+    min_x = np.nanmin(points[:, 0])
+    max_x = np.nanmax(points[:, 0])
+    min_y = np.nanmin(points[:, 1])
+    max_y = np.nanmax(points[:, 1])
 
     extent = (
         np.round(min_x - 0.05 * (max_x - min_x)),
@@ -263,6 +261,7 @@ def _datashade_points(
     width=800,
     height=800,
     show_legend=True,
+    alpha=255,
 ):
 
     """Use datashader to plot points"""
@@ -290,7 +289,7 @@ def _datashade_points(
         data["label"] = pd.Categorical(labels)
         aggregation = canvas.points(data, "x", "y", agg=ds.count_cat("label"))
         if color_key is None and color_key_cmap is None:
-            result = tf.shade(aggregation, how="eq_hist")
+            result = tf.shade(aggregation, how="eq_hist", alpha=alpha)
         elif color_key is None:
             unique_labels = np.unique(labels)
             num_labels = unique_labels.shape[0]
@@ -301,12 +300,16 @@ def _datashade_points(
                 Patch(facecolor=color_key[i], label=k)
                 for i, k in enumerate(unique_labels)
             ]
-            result = tf.shade(aggregation, color_key=color_key, how="eq_hist")
+            result = tf.shade(
+                aggregation, color_key=color_key, how="eq_hist", alpha=alpha
+            )
         else:
             legend_elements = [
                 Patch(facecolor=color_key[k], label=k) for k in color_key.keys()
             ]
-            result = tf.shade(aggregation, color_key=color_key, how="eq_hist")
+            result = tf.shade(
+                aggregation, color_key=color_key, how="eq_hist", alpha=alpha
+            )
 
     # Color by values
     elif values is not None:
@@ -326,7 +329,9 @@ def _datashade_points(
             )
             aggregation = canvas.points(data, "x", "y", agg=ds.count_cat("val_cat"))
             color_key = _to_hex(plt.get_cmap(cmap)(np.linspace(0, 1, 256)))
-            result = tf.shade(aggregation, color_key=color_key, how="eq_hist")
+            result = tf.shade(
+                aggregation, color_key=color_key, how="eq_hist", alpha=alpha
+            )
         else:
             data["val_cat"] = pd.Categorical(values)
             aggregation = canvas.points(data, "x", "y", agg=ds.count_cat("val_cat"))
@@ -334,12 +339,14 @@ def _datashade_points(
                 plt.get_cmap(cmap)(np.linspace(0, 1, unique_values.shape[0]))
             )
             color_key = dict(zip(unique_values, color_key_cols))
-            result = tf.shade(aggregation, color_key=color_key, how="eq_hist")
+            result = tf.shade(
+                aggregation, color_key=color_key, how="eq_hist", alpha=alpha
+            )
 
     # Color by density (default datashader option)
     else:
         aggregation = canvas.points(data, "x", "y", agg=ds.count())
-        result = tf.shade(aggregation, cmap=plt.get_cmap(cmap))
+        result = tf.shade(aggregation, cmap=plt.get_cmap(cmap), alpha=alpha)
 
     if background is not None:
         result = tf.set_background(result, background)
@@ -365,6 +372,7 @@ def _matplotlib_points(
     width=800,
     height=800,
     show_legend=True,
+    alpha=None,
 ):
     """Use matplotlib to plot points"""
     point_size = 100.0 / np.sqrt(points.shape[0])
@@ -409,14 +417,17 @@ def _matplotlib_points(
                     "Color key must have enough colors for the number of labels"
                 )
 
-            new_color_key = {k: color_key[i] for i, k in enumerate(unique_labels)}
+            new_color_key = {
+                k: matplotlib.colors.to_hex(color_key[i])
+                for i, k in enumerate(unique_labels)
+            }
             legend_elements = [
                 Patch(facecolor=color_key[i], label=k)
                 for i, k in enumerate(unique_labels)
             ]
             colors = pd.Series(labels).map(new_color_key)
 
-        ax.scatter(points[:, 0], points[:, 1], s=point_size, c=colors)
+        ax.scatter(points[:, 0], points[:, 1], s=point_size, c=colors, alpha=alpha)
 
     # Color by values
     elif values is not None:
@@ -427,7 +438,9 @@ def _matplotlib_points(
                     values.shape[0], points.shape[0]
                 )
             )
-        ax.scatter(points[:, 0], points[:, 1], s=point_size, c=values, cmap=cmap)
+        ax.scatter(
+            points[:, 0], points[:, 1], s=point_size, c=values, cmap=cmap, alpha=alpha
+        )
 
     # No color (just pick the midpoint of the cmap)
     else:
@@ -455,8 +468,10 @@ def show(plot_to_show):
     """
     if isinstance(plot_to_show, plt.Axes):
         show_static()
-    elif isinstance(plot_to_show, bpl.Figure):
+    elif isinstance(plot_to_show, bpl.figure):
         show_interactive(plot_to_show)
+    elif isinstance(plot_to_show, hv.core.spaces.DynamicMap):
+        show_interactive(hv.render(plot_to_show), backend="bokeh")
     else:
         raise ValueError(
             "The type of ``plot_to_show`` was not valid, or not understood."
@@ -465,6 +480,7 @@ def show(plot_to_show):
 
 def points(
     umap_object,
+    points=None,
     labels=None,
     values=None,
     theme=None,
@@ -476,13 +492,15 @@ def points(
     height=800,
     show_legend=True,
     subset_points=None,
+    ax=None,
+    alpha=None,
 ):
     """Plot an embedding as points. Currently this only works
     for 2D embeddings. While there are many optional parameters
     to further control and tailor the plotting, you need only
     pass in the trained/fit umap model to get results. This plot
     utility will attempt to do the hard work of avoiding
-    overplotting issues, and make it easy to automatically
+    over-plotting issues, and make it easy to automatically
     colour points by a categorical labelling or numeric values.
 
     This method is intended to be used within a Jupyter
@@ -493,6 +511,12 @@ def points(
     umap_object: trained UMAP object
         A trained UMAP object that has a 2D embedding.
 
+    points: array, shape (n_samples, dim) (optional, default None)
+        An array of points to be plotted. Usually this is None
+        and so the original embedding points of the umap_object
+        are used. However points can be passed explicitly instead
+        which is useful for points manually transformed.
+
     labels: array, shape (n_samples,) (optional, default None)
         An array of labels (assumed integer or categorical),
         one for each data sample.
@@ -575,17 +599,24 @@ def points(
         A way to select a subset of points based on an array of boolean
         values.
 
+    ax: matplotlib axis (optional, default None)
+        The matplotlib axis to draw the plot to, or if None, which is
+        the default, a new axis will be created and returned.
+
+    alpha: float (optional, default: None)
+        The alpha blending value, between 0 (transparent) and 1 (opaque).
+
     Returns
     -------
     result: matplotlib axis
         The result is a matplotlib axis with the relevant plot displayed.
-        If you are using a notbooks and have ``%matplotlib inline`` set
+        If you are using a notebooks and have ``%matplotlib inline`` set
         then this will simply display inline.
     """
-    if not hasattr(umap_object, "embedding_"):
-        raise ValueError(
-            "UMAP object must perform fit on data before it can be visualized"
-        )
+    # if not hasattr(umap_object, "embedding_"):
+    #     raise ValueError(
+    #         "UMAP object must perform fit on data before it can be visualized"
+    #     )
 
     if theme is not None:
         cmap = _themes[theme]["cmap"]
@@ -597,7 +628,12 @@ def points(
             "Conflicting options; only one of labels or values should be set"
         )
 
-    points = umap_object.embedding_
+    if alpha is not None:
+        if not 0.0 <= alpha <= 1.0:
+            raise ValueError("Alpha must be between 0 and 1 inclusive")
+
+    if points is None:
+        points = _get_embedding(umap_object)
 
     if subset_points is not None:
         if len(subset_points) != points.shape[0]:
@@ -618,9 +654,10 @@ def points(
 
     font_color = _select_font_color(background)
 
-    dpi = plt.rcParams["figure.dpi"]
-    fig = plt.figure(figsize=(width / dpi, height / dpi))
-    ax = fig.add_subplot(111)
+    if ax is None:
+        dpi = plt.rcParams["figure.dpi"]
+        fig = plt.figure(figsize=(width / dpi, height / dpi))
+        ax = fig.add_subplot(111)
 
     if points.shape[0] <= width * height // 10:
         ax = _matplotlib_points(
@@ -635,8 +672,15 @@ def points(
             width,
             height,
             show_legend,
+            alpha,
         )
     else:
+        # Datashader uses 0-255 as the range for alpha, with 255 as the default
+        if alpha is not None:
+            alpha = alpha * 255
+        else:
+            alpha = 255
+
         ax = _datashade_points(
             points,
             ax,
@@ -649,15 +693,16 @@ def points(
             width,
             height,
             show_legend,
+            alpha,
         )
 
     ax.set(xticks=[], yticks=[])
-    if umap_object.metric != "euclidean":
+    if _get_metric(umap_object) != "euclidean":
         ax.text(
             0.99,
             0.01,
             "UMAP: metric={}, n_neighbors={}, min_dist={}".format(
-                umap_object.metric, umap_object.n_neighbors, umap_object.min_dist
+                _get_metric(umap_object), umap_object.n_neighbors, umap_object.min_dist
             ),
             transform=ax.transAxes,
             horizontalalignment="right",
@@ -702,7 +747,7 @@ def connectivity(
     to further control and tailor the plotting, you need only
     pass in the trained/fit umap model to get results. This plot
     utility will attempt to do the hard work of avoiding
-    overplotting issues and provide options for plotting the
+    over-plotting issues and provide options for plotting the
     points as well as using edge bundling for graph visualization.
 
     Parameters
@@ -804,7 +849,7 @@ def connectivity(
     -------
     result: matplotlib axis
         The result is a matplotlib axis with the relevant plot displayed.
-        If you are using a notbooks and have ``%matplotlib inline`` set
+        If you are using a notebook and have ``%matplotlib inline`` set
         then this will simply display inline.
     """
     if theme is not None:
@@ -813,7 +858,7 @@ def connectivity(
         edge_cmap = _themes[theme]["edge_cmap"]
         background = _themes[theme]["background"]
 
-    points = umap_object.embedding_
+    points = _get_embedding(umap_object)
     point_df = pd.DataFrame(points, columns=("x", "y"))
 
     point_size = 100.0 / np.sqrt(points.shape[0])
@@ -965,7 +1010,7 @@ def diagnostic(
         as that which provides ``local_variance_threshold``
         or more of the ``variance_explained_ratio``.
 
-    ax: matlotlib axis (optional, default None)
+    ax: matplotlib axis (optional, default None)
         A matplotlib axis to plot to, or, if None, a new
         axis will be created and returned.
 
@@ -983,11 +1028,11 @@ def diagnostic(
     -------
     result: matplotlib axis
         The result is a matplotlib axis with the relevant plot displayed.
-        If you are using a notbooks and have ``%matplotlib inline`` set
+        If you are using a notebook and have ``%matplotlib inline`` set
         then this will simply display inline.
     """
 
-    points = umap_object.embedding_
+    points = _get_embedding(umap_object)
 
     if points.shape[1] != 2:
         raise ValueError("Plotting is currently only implemented for 2D embeddings")
@@ -1176,6 +1221,7 @@ def interactive(
     labels=None,
     values=None,
     hover_data=None,
+    tools=None,
     theme=None,
     cmap="Blues",
     color_key=None,
@@ -1188,11 +1234,12 @@ def interactive(
     interactive_text_search=False,
     interactive_text_search_columns=None,
     interactive_text_search_alpha_contrast=0.95,
+    alpha=None,
 ):
     """Create an interactive bokeh plot of a UMAP embedding.
     While static plots are useful, sometimes a plot that
     supports interactive zooming, and hover tooltips for
-    individual points is much more desireable. This function
+    individual points is much more desirable. This function
     provides a simple interface for creating such plots. The
     result is a bokeh plot that will be displayed in a notebook.
 
@@ -1229,6 +1276,14 @@ def interactive(
         for each data point. Column names will be used for
         identifying information within the tooltip.
 
+    tools: List (optional, default None),
+        Defines the tools to be configured for interactive plots.
+        The list can be mixed type of string and tools objects defined by
+        Bokeh like HoverTool. Default tool list Bokeh uses is
+        ["pan","wheel_zoom","box_zoom","save","reset","help",].
+        When tools are specified, and includes hovertool, automatic tooltip
+        based on hover_data is not created.
+
     theme: string (optional, default None)
         A color theme to use for plotting. A small set of
         predefined themes are provided which have relatively
@@ -1304,6 +1359,9 @@ def interactive(
         Alpha value for points matching text search. Alpha value for points
         not matching text search will be 1 - interactive_text_search_alpha_contrast
 
+    alpha: float (optional, default: None)
+        The alpha blending value, between 0 (transparent) and 1 (opaque).
+
     Returns
     -------
 
@@ -1318,7 +1376,11 @@ def interactive(
             "Conflicting options; only one of labels or values should be set"
         )
 
-    points = umap_object.embedding_
+    if alpha is not None:
+        if not 0.0 <= alpha <= 1.0:
+            raise ValueError("Alpha must be between 0 and 1 inclusive")
+
+    points = _get_embedding(umap_object)
     if subset_points is not None:
         if len(subset_points) != points.shape[0]:
             raise ValueError(
@@ -1334,7 +1396,7 @@ def interactive(
     if point_size is None:
         point_size = 100.0 / np.sqrt(points.shape[0])
 
-    data = pd.DataFrame(umap_object.embedding_, columns=("x", "y"))
+    data = pd.DataFrame(_get_embedding(umap_object), columns=("x", "y"))
 
     if labels is not None:
         data["label"] = labels
@@ -1376,17 +1438,24 @@ def interactive(
             hover_data = hover_data[subset_points]
 
     if points.shape[0] <= width * height // 10:
-
+        tooltips = None
+        tooltip_needed = True
         if hover_data is not None:
             tooltip_dict = {}
             for col_name in hover_data:
                 data[col_name] = hover_data[col_name]
-                tooltip_dict[col_name] = "@" + col_name
+                tooltip_dict[col_name] = "@{" + col_name + "}"
             tooltips = list(tooltip_dict.items())
-        else:
-            tooltips = None
 
-        data["alpha"] = 1
+            for _tool in tools:
+                if _tool.__class__.__name__ == "HoverTool":
+                    tooltip_needed = False
+                    break
+
+        if alpha is not None:
+            data["alpha"] = alpha
+        else:
+            data["alpha"] = 1
 
         # bpl.output_notebook(hide_banner=True) # this doesn't work for non-notebook use
         data_source = bpl.ColumnDataSource(data)
@@ -1394,7 +1463,8 @@ def interactive(
         plot = bpl.figure(
             width=width,
             height=height,
-            tooltips=tooltips,
+            tooltips=None if not tooltip_needed else tooltips,
+            tools=tools if tools is not None else "pan,wheel_zoom,box_zoom,save,reset,help",
             background_fill_color=background,
         )
         plot.circle(
@@ -1471,15 +1541,18 @@ def interactive(
         if hover_data is not None:
             warn(
                 "Too many points for hover data -- tooltips will not"
-                "be displayed. Sorry; try subssampling your data."
+                "be displayed. Sorry; try subsampling your data."
             )
         if interactive_text_search:
             warn(
-                "Too many points for text search." "Sorry; try subssampling your data."
+                "Too many points for text search." "Sorry; try subsampling your data."
             )
+        if alpha is not None:
+            warn("Alpha parameter will not be applied on holoviews plots")
         hv.extension("bokeh")
         hv.output(size=300)
-        hv.opts('RGB [bgcolor="{}", xaxis=None, yaxis=None]'.format(background))
+        hv.opts.defaults(hv.opts.RGB(bgcolor=background, xaxis=None, yaxis=None))
+
         if labels is not None:
             point_plot = hv.Points(data, kdims=["x", "y"])
             plot = hd.datashade(
@@ -1515,3 +1588,37 @@ def interactive(
             )
 
     return plot
+
+
+def nearest_neighbour_distribution(umap_object, bins=25, ax=None):
+    """Create a histogram of the average distance to each points
+    nearest neighbors.
+
+    Parameters
+    ----------
+    umap_object: trained UMAP object
+        A trained UMAP object that has an embedding.
+
+    bins: int (optional, default 25)
+        Number of bins to put the points into
+
+    ax: matplotlib axis (optional, default None)
+        A matplotlib axis to plot to, or, if None, a new
+        axis will be created and returned.
+
+    Returns
+    -------
+
+    """
+    nn_distances = average_nn_distance(umap_object.graph_)
+
+    if ax is None:
+        fig = plt.figure()
+        ax = fig.add_subplot(111)
+
+    ax.set_xlabel(f"Average distance to nearest neighbors")
+    ax.set_ylabel("Frequency")
+
+    ax.hist(nn_distances, bins=bins)
+
+    return ax
