@@ -47,7 +47,7 @@ def test_blobs_cluster():
 # Multi-components Layout
 def test_multi_component_layout():
     data, labels = make_blobs(
-        100, 2, centers=5, cluster_std=0.5, center_box=[-20, 20], random_state=42
+        100, 2, centers=5, cluster_std=0.5, center_box=(-20, 20), random_state=42
     )
 
     true_centroids = np.empty((labels.max() + 1, data.shape[1]), dtype=np.float64)
@@ -74,7 +74,7 @@ def test_multi_component_layout():
 # Multi-components Layout
 def test_multi_component_layout_precomputed():
     data, labels = make_blobs(
-        100, 2, centers=5, cluster_std=0.5, center_box=[-20, 20], random_state=42
+        100, 2, centers=5, cluster_std=0.5, center_box=(-20, 20), random_state=42
     )
     dmat = pairwise_distances(data)
 
@@ -259,7 +259,7 @@ def test_umap_update_large(
 
     error = np.sum(np.abs((new_model.graph_ - comparison_graph).data))
 
-    assert error < 3.0 # Higher error tolerance based on approx nearest neighbors
+    assert error < 3.0  # Higher error tolerance based on approx nearest neighbors
 
 
 # -----------------
@@ -292,7 +292,7 @@ def test_component_layout_options(nn_data):
         n_components,
         component_labels,
         2,
-        np.random,
+        None,
         metric="precomputed",
         metric_kwds={"linkage": "single"},
     )
@@ -301,7 +301,7 @@ def test_component_layout_options(nn_data):
         n_components,
         component_labels,
         2,
-        np.random,
+        None,
         metric="precomputed",
         metric_kwds={"linkage": "average"},
     )
@@ -310,7 +310,7 @@ def test_component_layout_options(nn_data):
         n_components,
         component_labels,
         2,
-        np.random,
+        None,
         metric="precomputed",
         metric_kwds={"linkage": "complete"},
     )
